@@ -1,19 +1,13 @@
 package consul
 
 import (
-	"context"
 	"errors"
 	"fmt"
 	"reflect"
-	"strings"
-	"sync"
 	"time"
 
-	"github.com/hashicorp/consul/lib/semaphore"
 	"github.com/hashicorp/go-hclog"
 
-	"golang.org/x/time/rate"
-
 	"github.com/hashicorp/consul/acl"
 	"github.com/hashicorp/consul/agent/connect"
 	"github.com/hashicorp/consul/agent/connect/ca"
@@ -53,57 +47,6 @@ type ConnectCA struct {
 	srv *Server
 
 	logger hclog.Logger
-
-	// csrRateLimiter limits the rate of signing new certs if configured. Lazily
-	// initialized from current config to support dynamic changes.
-	// csrRateLimiterMu must be held while dereferencing the pointer or storing a
-	// new one, but methods can be called on the limiter object outside of the
-	// locked section. This is done only in the getCSRRateLimiterWithLimit method.
-	csrRateLimiter   *rate.Limiter
-	csrRateLimiterMu sync.RWMutex
-
-	// csrConcurrencyLimiter is a dynamically resizable semaphore used to limit
-	// Sign RPC concurrency if configured. The zero value is usable as soon as
-	// SetSize is called which we do dynamically in the RPC handler to avoid
-	// having to hook elaborate synchronization mechanisms through the CA config
-	// endpoint and config reload etc.
-	csrConcurrencyLimiter semaphore.Dynamic
-}
-
-// getCSRRateLimiterWithLimit returns a rate.Limiter with the desired limit set.
-// It uses the shared server-wide limiter unless the limit has been changed in
-// config or the limiter has not been setup yet in which case it just-in-time
-// configures the new limiter. We assume that limit changes are relatively rare
-// and that all callers (there is currently only one) use the same config value
-// as the limit. There might be some flapping if there are multiple concurrent
-// requests in flight at the time the config changes where A sees the new value
-// and updates, B sees the old but then gets this lock second and changes back.
-// Eventually though and very soon (once all current RPCs are complete) we are
-// guaranteed to have the correct limit set by the next RPC that comes in so I
-// assume this is fine. If we observe strange behavior because of it, we could
-// add hysteresis that prevents changes too soon after a previous change but
-// that seems unnecessary for now.
-func (s *ConnectCA) getCSRRateLimiterWithLimit(limit rate.Limit) *rate.Limiter {
-	s.csrRateLimiterMu.RLock()
-	lim := s.csrRateLimiter
-	s.csrRateLimiterMu.RUnlock()
-
-	// If there is a current limiter with the same limit, return it. This should
-	// be the common case.
-	if lim != nil && lim.Limit() == limit {
-		return lim
-	}
-
-	// Need to change limiter, get write lock
-	s.csrRateLimiterMu.Lock()
-	defer s.csrRateLimiterMu.Unlock()
-	// No limiter yet, or limit changed in CA config, reconfigure a new limiter.
-	// We use burst of 1 for a hard limit. Note that either bursting or waiting is
-	// necessary to get expected behavior in fact of random arrival times, but we
-	// don't need both and we use Wait with a small delay to smooth noise. See
-	// https://github.com/banks/sim-rate-limit-backoff/blob/master/README.md.
-	s.csrRateLimiter = rate.NewLimiter(limit, 1)
-	return s.csrRateLimiter
 }
 
 // ConfigurationGet returns the configuration for the CA.
@@ -115,7 +58,7 @@ func (s *ConnectCA) ConfigurationGet(
 		return ErrConnectNotEnabled
 	}
 
-	if done, err := s.srv.forward("ConnectCA.ConfigurationGet", args, args, reply); done {
+	if done, err := s.srv.ForwardRPC("ConnectCA.ConfigurationGet", args, args, reply); done {
 		return err
 	}
 
@@ -124,7 +67,7 @@ func (s *ConnectCA) ConfigurationGet(
 	if err != nil {
 		return err
 	}
-	if rule != nil && rule.OperatorRead(nil) != acl.Allow {
+	if rule != nil && rule.OperatorWrite(nil) != acl.Allow {
 		return acl.ErrPermissionDenied
 	}
 
@@ -147,7 +90,7 @@ func (s *ConnectCA) ConfigurationSet(
 		return ErrConnectNotEnabled
 	}
 
-	if done, err := s.srv.forward("ConnectCA.ConfigurationSet", args, args, reply); done {
+	if done, err := s.srv.ForwardRPC("ConnectCA.ConfigurationSet", args, args, reply); done {
 		return err
 	}
 
@@ -209,6 +152,17 @@ func (s *ConnectCA) ConfigurationSet(
 	if err := newProvider.Configure(pCfg); err != nil {
 		return fmt.Errorf("error configuring provider: %v", err)
 	}
+
+	// Set up a defer to clean up the new provider if we exit early due to an error.
+	cleanupNewProvider := true
+	defer func() {
+		if cleanupNewProvider {
+			if err := newProvider.Cleanup(); err != nil {
+				s.logger.Warn("failed to clean up CA provider while handling startup failure", "provider", newProvider, "error", err)
+			}
+		}
+	}()
+
 	if err := newProvider.GenerateRoot(); err != nil {
 		return fmt.Errorf("error generating CA root certificate: %v", err)
 	}
@@ -252,6 +206,7 @@ func (s *ConnectCA) ConfigurationSet(
 		}
 
 		// If the config has been committed, update the local provider instance
+		cleanupNewProvider = false
 		s.srv.setCAProvider(newProvider, newActiveRoot)
 
 		s.logger.Info("CA provider config updated")
@@ -348,6 +303,7 @@ func (s *ConnectCA) ConfigurationSet(
 
 	// If the config has been committed, update the local provider instance
 	// and call teardown on the old provider
+	cleanupNewProvider = false
 	s.srv.setCAProvider(newProvider, newActiveRoot)
 
 	if err := oldProvider.Cleanup(); err != nil {
@@ -364,7 +320,7 @@ func (s *ConnectCA) Roots(
 	args *structs.DCSpecificRequest,
 	reply *structs.IndexedCARoots) error {
 	// Forward if necessary
-	if done, err := s.srv.forward("ConnectCA.Roots", args, args, reply); done {
+	if done, err := s.srv.ForwardRPC("ConnectCA.Roots", args, args, reply); done {
 		return err
 	}
 
@@ -376,55 +332,12 @@ func (s *ConnectCA) Roots(
 	return s.srv.blockingQuery(
 		&args.QueryOptions, &reply.QueryMeta,
 		func(ws memdb.WatchSet, state *state.Store) error {
-			index, roots, config, err := state.CARootsAndConfig(ws)
+			roots, err := s.srv.getCARoots(ws, state)
 			if err != nil {
 				return err
 			}
 
-			if config != nil {
-				// Build TrustDomain based on the ClusterID stored.
-				signingID := connect.SpiffeIDSigningForCluster(config)
-				if signingID == nil {
-					// If CA is bootstrapped at all then this should never happen but be
-					// defensive.
-					return errors.New("no cluster trust domain setup")
-				}
-				reply.TrustDomain = signingID.Host()
-			}
-
-			reply.Index, reply.Roots = index, roots
-			if reply.Roots == nil {
-				reply.Roots = make(structs.CARoots, 0)
-			}
-
-			// The API response must NEVER contain the secret information
-			// such as keys and so on. We use a whitelist below to copy the
-			// specific fields we want to expose.
-			for i, r := range reply.Roots {
-				// IMPORTANT: r must NEVER be modified, since it is a pointer
-				// directly to the structure in the memdb store.
-
-				reply.Roots[i] = &structs.CARoot{
-					ID:                  r.ID,
-					Name:                r.Name,
-					SerialNumber:        r.SerialNumber,
-					SigningKeyID:        r.SigningKeyID,
-					ExternalTrustDomain: r.ExternalTrustDomain,
-					NotBefore:           r.NotBefore,
-					NotAfter:            r.NotAfter,
-					RootCert:            r.RootCert,
-					IntermediateCerts:   r.IntermediateCerts,
-					RaftIndex:           r.RaftIndex,
-					Active:              r.Active,
-					PrivateKeyType:      r.PrivateKeyType,
-					PrivateKeyBits:      r.PrivateKeyBits,
-				}
-
-				if r.Active {
-					reply.ActiveRootID = r.ID
-				}
-			}
-
+			*reply = *roots
 			return nil
 		},
 	)
@@ -439,7 +352,7 @@ func (s *ConnectCA) Sign(
 		return ErrConnectNotEnabled
 	}
 
-	if done, err := s.srv.forward("ConnectCA.Sign", args, args, reply); done {
+	if done, err := s.srv.ForwardRPC("ConnectCA.Sign", args, args, reply); done {
 		return err
 	}
 
@@ -455,40 +368,21 @@ func (s *ConnectCA) Sign(
 		return err
 	}
 
-	provider, caRoot := s.srv.getCAProvider()
-	if provider == nil {
-		return fmt.Errorf("internal error: CA provider is nil")
-	} else if caRoot == nil {
-		return fmt.Errorf("internal error: CA root is nil")
-	}
-
-	// Verify that the CSR entity is in the cluster's trust domain
-	state := s.srv.fsm.State()
-	_, config, err := state.CAConfig(nil)
+	// Verify that the ACL token provided has permission to act as this service
+	rule, err := s.srv.ResolveToken(args.Token)
 	if err != nil {
 		return err
 	}
-	signingID := connect.SpiffeIDSigningForCluster(config)
+
+	var authzContext acl.AuthorizerContext
+	var entMeta structs.EnterpriseMeta
+
 	serviceID, isService := spiffeID.(*connect.SpiffeIDService)
 	agentID, isAgent := spiffeID.(*connect.SpiffeIDAgent)
 	if !isService && !isAgent {
 		return fmt.Errorf("SPIFFE ID in CSR must be a service or agent ID")
 	}
 
-	if isService {
-		if !signingID.CanSign(spiffeID) {
-			return fmt.Errorf("SPIFFE ID in CSR from a different trust domain: %s, "+
-				"we are %s", serviceID.Host, signingID.Host())
-		}
-	}
-
-	// Verify that the ACL token provided has permission to act as this service
-	rule, err := s.srv.ResolveToken(args.Token)
-	if err != nil {
-		return err
-	}
-	var authzContext acl.AuthorizerContext
-	var entMeta structs.EnterpriseMeta
 	if isService {
 		entMeta.Merge(serviceID.GetEnterpriseMeta())
 		entMeta.FillAuthzContext(&authzContext)
@@ -509,106 +403,11 @@ func (s *ConnectCA) Sign(
 		}
 	}
 
-	commonCfg, err := config.GetCommonConfig()
+	cert, err := s.srv.SignCertificate(csr, spiffeID)
 	if err != nil {
 		return err
 	}
-	if commonCfg.CSRMaxPerSecond > 0 {
-		lim := s.getCSRRateLimiterWithLimit(rate.Limit(commonCfg.CSRMaxPerSecond))
-		// Wait up to the small threshold we allow for a token.
-		ctx, cancel := context.WithTimeout(context.Background(), csrLimitWait)
-		defer cancel()
-		if lim.Wait(ctx) != nil {
-			return ErrRateLimited
-		}
-	} else if commonCfg.CSRMaxConcurrent > 0 {
-		s.csrConcurrencyLimiter.SetSize(int64(commonCfg.CSRMaxConcurrent))
-		ctx, cancel := context.WithTimeout(context.Background(), csrLimitWait)
-		defer cancel()
-		if err := s.csrConcurrencyLimiter.Acquire(ctx); err != nil {
-			return ErrRateLimited
-		}
-		defer s.csrConcurrencyLimiter.Release()
-	}
-
-	// All seems to be in order, actually sign it.
-	pem, err := provider.Sign(csr)
-	if err == ca.ErrRateLimited {
-		return ErrRateLimited
-	}
-	if err != nil {
-		return err
-	}
-
-	// Append any intermediates needed by this root.
-	for _, p := range caRoot.IntermediateCerts {
-		pem = strings.TrimSpace(pem) + "\n" + p
-	}
-
-	// Append our local CA's intermediate if there is one.
-	inter, err := provider.ActiveIntermediate()
-	if err != nil {
-		return err
-	}
-	root, err := provider.ActiveRoot()
-	if err != nil {
-		return err
-	}
-
-	if inter != root {
-		pem = strings.TrimSpace(pem) + "\n" + inter
-	}
-
-	// TODO(banks): when we implement IssuedCerts table we can use the insert to
-	// that as the raft index to return in response.
-	//
-	// UPDATE(mkeeler): The original implementation relied on updating the CAConfig
-	// and using its index as the ModifyIndex for certs. This was buggy. The long
-	// term goal is still to insert some metadata into raft about the certificates
-	// and use that raft index for the ModifyIndex. This is a partial step in that
-	// direction except that we only are setting an index and not storing the
-	// metadata.
-	req := structs.CALeafRequest{
-		Op:           structs.CALeafOpIncrementIndex,
-		Datacenter:   s.srv.config.Datacenter,
-		WriteRequest: structs.WriteRequest{Token: args.Token},
-	}
-
-	resp, err := s.srv.raftApply(structs.ConnectCALeafRequestType|structs.IgnoreUnknownTypeFlag, &req)
-	if err != nil {
-		return err
-	}
-
-	modIdx, ok := resp.(uint64)
-	if !ok {
-		return fmt.Errorf("Invalid response from updating the leaf cert index")
-	}
-
-	cert, err := connect.ParseCert(pem)
-	if err != nil {
-		return err
-	}
-
-	// Set the response
-	*reply = structs.IssuedCert{
-		SerialNumber:   connect.EncodeSerialNumber(cert.SerialNumber),
-		CertPEM:        pem,
-		ValidAfter:     cert.NotBefore,
-		ValidBefore:    cert.NotAfter,
-		EnterpriseMeta: entMeta,
-		RaftIndex: structs.RaftIndex{
-			ModifyIndex: modIdx,
-			CreateIndex: modIdx,
-		},
-	}
-	if isService {
-		reply.Service = serviceID.Service
-		reply.ServiceURI = cert.URIs[0].String()
-	} else if isAgent {
-		reply.Agent = agentID.Agent
-		reply.AgentURI = cert.URIs[0].String()
-	}
-
+	*reply = *cert
 	return nil
 }
 
@@ -621,7 +420,7 @@ func (s *ConnectCA) SignIntermediate(
 		return ErrConnectNotEnabled
 	}
 
-	if done, err := s.srv.forward("ConnectCA.SignIntermediate", args, args, reply); done {
+	if done, err := s.srv.ForwardRPC("ConnectCA.SignIntermediate", args, args, reply); done {
 		return err
 	}
 
