@@ -1,6 +1,7 @@
 package consul
 
 import (
+	"context"
 	"errors"
 	"fmt"
 	"io"
@@ -20,6 +21,7 @@ import (
 	"github.com/hashicorp/consul/acl"
 	ca "github.com/hashicorp/consul/agent/connect/ca"
 	"github.com/hashicorp/consul/agent/consul/authmethod"
+	"github.com/hashicorp/consul/agent/consul/authmethod/ssoauth"
 	"github.com/hashicorp/consul/agent/consul/autopilot"
 	"github.com/hashicorp/consul/agent/consul/fsm"
 	"github.com/hashicorp/consul/agent/consul/state"
@@ -35,6 +37,7 @@ import (
 	connlimit "github.com/hashicorp/go-connlimit"
 	"github.com/hashicorp/go-hclog"
 	"github.com/hashicorp/go-memdb"
+	"github.com/hashicorp/memberlist"
 	"github.com/hashicorp/raft"
 	raftboltdb "github.com/hashicorp/raft-boltdb"
 	"github.com/hashicorp/serf/serf"
@@ -45,6 +48,8 @@ import (
 // Consul-level protocol versions, that are used to configure the Serf
 // protocol versions.
 const (
+	DefaultRPCProtocol = 2
+
 	ProtocolVersionMin uint8 = 2
 
 	// Version 3 added support for network coordinates but we kept the
@@ -90,17 +95,20 @@ const (
 )
 
 const (
-	legacyACLReplicationRoutineName    = "legacy ACL replication"
-	aclPolicyReplicationRoutineName    = "ACL policy replication"
-	aclRoleReplicationRoutineName      = "ACL role replication"
-	aclTokenReplicationRoutineName     = "ACL token replication"
-	aclTokenReapingRoutineName         = "acl token reaping"
-	aclUpgradeRoutineName              = "legacy ACL token upgrade"
-	caRootPruningRoutineName           = "CA root pruning"
-	configReplicationRoutineName       = "config entry replication"
-	intentionReplicationRoutineName    = "intention replication"
-	secondaryCARootWatchRoutineName    = "secondary CA roots watch"
-	secondaryCertRenewWatchRoutineName = "secondary cert renew watch"
+	legacyACLReplicationRoutineName       = "legacy ACL replication"
+	aclPolicyReplicationRoutineName       = "ACL policy replication"
+	aclRoleReplicationRoutineName         = "ACL role replication"
+	aclTokenReplicationRoutineName        = "ACL token replication"
+	aclTokenReapingRoutineName            = "acl token reaping"
+	aclUpgradeRoutineName                 = "legacy ACL token upgrade"
+	caRootPruningRoutineName              = "CA root pruning"
+	configReplicationRoutineName          = "config entry replication"
+	federationStateReplicationRoutineName = "federation state replication"
+	federationStateAntiEntropyRoutineName = "federation state anti-entropy"
+	federationStatePruningRoutineName     = "federation state pruning"
+	intentionReplicationRoutineName       = "intention replication"
+	secondaryCARootWatchRoutineName       = "secondary CA roots watch"
+	intermediateCertRenewWatchRoutineName = "intermediate cert renew watch"
 )
 
 var (
@@ -146,6 +154,9 @@ type Server struct {
 	caProviderRoot *structs.CARoot
 	caProviderLock sync.RWMutex
 
+	// rate limiter to use when signing leaf certificates
+	caLeafLimiter connectSignRateLimiter
+
 	// Consul configuration
 	config *Config
 
@@ -153,6 +164,16 @@ type Server struct {
 	// centralized config
 	configReplicator *Replicator
 
+	// federationStateReplicator is used to manage the leaders replication routines for
+	// federation states
+	federationStateReplicator *Replicator
+
+	// dcSupportsFederationStates is used to determine whether we can
+	// replicate federation states or not. All servers in the local
+	// DC must be on a version of Consul supporting federation states
+	// before this will get enabled.
+	dcSupportsFederationStates int32
+
 	// tokens holds ACL tokens initially from the configuration, but can
 	// be updated at runtime, so should always be used instead of going to
 	// the configuration directly.
@@ -240,7 +261,9 @@ type Server struct {
 
 	// serfWAN is the Serf cluster maintained between DC's
 	// which SHOULD only consist of Consul servers
-	serfWAN *serf.Serf
+	serfWAN                *serf.Serf
+	memberlistTransportWAN memberlist.IngestionAwareTransport
+	gatewayLocator         *GatewayLocator
 
 	// serverLookup tracks server consuls in the local datacenter.
 	// Used to do leader forwarding and provide fast lookup by server id and address
@@ -290,44 +313,28 @@ type Server struct {
 	EnterpriseServer
 }
 
-// NewServer is only used to help setting up a server for testing. Normal code
-// exercises NewServerLogger.
-func NewServer(config *Config) (*Server, error) {
-	c, err := tlsutil.NewConfigurator(config.ToTLSUtilConfig(), nil)
-	if err != nil {
-		return nil, err
-	}
-	return NewServerLogger(config, nil, new(token.Store), c)
-}
+// NewServer is used to construct a new Consul server from the configuration
+// and extra options, potentially returning an error.
+func NewServer(config *Config, options ...ConsulOption) (*Server, error) {
+	flat := flattenConsulOptions(options)
+
+	logger := flat.logger
+	tokens := flat.tokens
+	tlsConfigurator := flat.tlsConfigurator
+	connPool := flat.connPool
+	rpcRouter := flat.router
 
-// NewServerLogger is used to construct a new Consul server from the
-// configuration, potentially returning an error
-func NewServerLogger(config *Config, logger hclog.InterceptLogger, tokens *token.Store, tlsConfigurator *tlsutil.Configurator) (*Server, error) {
-	// Check the protocol version.
 	if err := config.CheckProtocolVersion(); err != nil {
 		return nil, err
 	}
-
-	// Check for a data directory.
 	if config.DataDir == "" && !config.DevMode {
 		return nil, fmt.Errorf("Config must provide a DataDir")
 	}
-
-	// Sanity check the ACLs.
 	if err := config.CheckACL(); err != nil {
 		return nil, err
 	}
-
-	// Ensure we have a log output and create a logger.
-	if config.LogOutput == nil {
-		config.LogOutput = os.Stderr
-	}
-
 	if logger == nil {
-		logger = hclog.NewInterceptLogger(&hclog.LoggerOptions{
-			Level:  hclog.Debug,
-			Output: config.LogOutput,
-		})
+		return nil, fmt.Errorf("logger is required")
 	}
 
 	// Check if TLS is enabled
@@ -357,17 +364,25 @@ func NewServerLogger(config *Config, logger hclog.InterceptLogger, tokens *token
 	// Create the shutdown channel - this is closed but never written to.
 	shutdownCh := make(chan struct{})
 
-	connPool := &pool.ConnPool{
-		SrcAddr:         config.RPCSrcAddr,
-		LogOutput:       config.LogOutput,
-		MaxTime:         serverRPCCache,
-		MaxStreams:      serverMaxStreams,
-		TLSConfigurator: tlsConfigurator,
-		ForceTLS:        config.VerifyOutgoing,
+	if connPool == nil {
+		connPool = &pool.ConnPool{
+			Server:          true,
+			SrcAddr:         config.RPCSrcAddr,
+			Logger:          logger.StandardLogger(&hclog.StandardLoggerOptions{InferLevels: true}),
+			MaxTime:         serverRPCCache,
+			MaxStreams:      serverMaxStreams,
+			TLSConfigurator: tlsConfigurator,
+			Datacenter:      config.Datacenter,
+		}
 	}
 
 	serverLogger := logger.NamedIntercept(logging.ConsulServer)
 	loggers := newLoggerStore(serverLogger)
+
+	if rpcRouter == nil {
+		rpcRouter = router.NewRouter(serverLogger, config.Datacenter, fmt.Sprintf("%s.%s", config.NodeName, config.Datacenter))
+	}
+
 	// Create server.
 	s := &Server{
 		config:                  config,
@@ -379,7 +394,7 @@ func NewServerLogger(config *Config, logger hclog.InterceptLogger, tokens *token
 		loggers:                 loggers,
 		leaveCh:                 make(chan struct{}),
 		reconcileCh:             make(chan serf.Member, reconcileChSize),
-		router:                  router.NewRouter(serverLogger, config.Datacenter, fmt.Sprintf("%s.%s", config.NodeName, config.Datacenter)),
+		router:                  rpcRouter,
 		rpcServer:               rpc.NewServer(),
 		insecureRPCServer:       rpc.NewServer(),
 		tlsConfigurator:         tlsConfigurator,
@@ -393,6 +408,16 @@ func NewServerLogger(config *Config, logger hclog.InterceptLogger, tokens *token
 		aclAuthMethodValidators: authmethod.NewCache(),
 	}
 
+	if s.config.ConnectMeshGatewayWANFederationEnabled {
+		s.gatewayLocator = NewGatewayLocator(
+			s.logger,
+			s,
+			s.config.Datacenter,
+			s.config.PrimaryDatacenter,
+		)
+		s.connPool.GatewayResolver = s.gatewayLocator.PickGateway
+	}
+
 	// Initialize enterprise specific server functionality
 	if err := s.initEnterprise(); err != nil {
 		s.Shutdown()
@@ -414,6 +439,26 @@ func NewServerLogger(config *Config, logger hclog.InterceptLogger, tokens *token
 		return nil, err
 	}
 
+	federationStateReplicatorConfig := ReplicatorConfig{
+		Name: logging.FederationState,
+		Delegate: &IndexReplicator{
+			Delegate: &FederationStateReplicator{
+				srv:            s,
+				gatewayLocator: s.gatewayLocator,
+			},
+			Logger: s.logger,
+		},
+		Rate:             s.config.FederationStateReplicationRate,
+		Burst:            s.config.FederationStateReplicationBurst,
+		Logger:           logger,
+		SuppressErrorLog: isErrFederationStatesNotSupported,
+	}
+	s.federationStateReplicator, err = NewReplicator(&federationStateReplicatorConfig)
+	if err != nil {
+		s.Shutdown()
+		return nil, err
+	}
+
 	// Initialize the stats fetcher that autopilot will use.
 	s.statsFetcher = NewStatsFetcher(logger, s.connPool, s.config.Datacenter)
 
@@ -452,8 +497,12 @@ func NewServerLogger(config *Config, logger hclog.InterceptLogger, tokens *token
 		return nil, fmt.Errorf("Failed to start Raft: %v", err)
 	}
 
-	if s.config.ConnectEnabled && s.config.AutoEncryptAllowTLS {
-		go s.trackAutoEncryptCARoots()
+	if s.config.ConnectEnabled && (s.config.AutoEncryptAllowTLS || s.config.AutoConfigAuthzEnabled) {
+		go s.connectCARootsMonitor(&lib.StopChannelContext{StopCh: s.shutdownCh})
+	}
+
+	if s.gatewayLocator != nil {
+		go s.gatewayLocator.Run(&lib.StopChannelContext{StopCh: s.shutdownCh})
 	}
 
 	// Serf and dynamic bind ports
@@ -474,6 +523,11 @@ func NewServerLogger(config *Config, logger hclog.InterceptLogger, tokens *token
 			s.Shutdown()
 			return nil, fmt.Errorf("Failed to start WAN Serf: %v", err)
 		}
+
+		// This is always a *memberlist.NetTransport or something which wraps
+		// it which satisfies this interface.
+		s.memberlistTransportWAN = config.SerfWANConfig.MemberlistConfig.Transport.(memberlist.IngestionAwareTransport)
+
 		// See big comment above why we are doing this.
 		if serfBindPortWAN == 0 {
 			serfBindPortWAN = config.SerfWANConfig.MemberlistConfig.BindPort
@@ -497,6 +551,11 @@ func NewServerLogger(config *Config, logger hclog.InterceptLogger, tokens *token
 		s.Shutdown()
 		return nil, fmt.Errorf("Failed to start LAN Serf: %v", err)
 	}
+
+	if err := s.router.AddArea(types.AreaLAN, s.serfLAN, s.connPool); err != nil {
+		s.Shutdown()
+		return nil, fmt.Errorf("Failed to add LAN serf route: %w", err)
+	}
 	go s.lanEventHandler()
 
 	// Start the flooders after the LAN event handler is wired up.
@@ -552,14 +611,8 @@ func NewServerLogger(config *Config, logger hclog.InterceptLogger, tokens *token
 	return s, nil
 }
 
-func (s *Server) trackAutoEncryptCARoots() {
+func (s *Server) connectCARootsMonitor(ctx context.Context) {
 	for {
-		select {
-		case <-s.shutdownCh:
-			s.logger.Debug("shutting down trackAutoEncryptCARoots because shutdown")
-			return
-		default:
-		}
 		ws := memdb.NewWatchSet()
 		state := s.fsm.State()
 		ws.Add(state.AbandonCh())
@@ -572,10 +625,14 @@ func (s *Server) trackAutoEncryptCARoots() {
 		for _, ca := range cas {
 			caPems = append(caPems, ca.RootCert)
 		}
-		if err := s.tlsConfigurator.UpdateAutoEncryptCA(caPems); err != nil {
+		if err := s.tlsConfigurator.UpdateAutoTLSCA(caPems); err != nil {
 			s.logger.Error("Failed to update AutoEncrypt CAPems", "error", err)
 		}
-		ws.Watch(nil)
+
+		if err := ws.WatchCtx(ctx); err == context.Canceled {
+			s.logger.Info("shutting down Connect CA roots monitor")
+			return
+		}
 	}
 }
 
@@ -656,7 +713,7 @@ func (s *Server) setupRaft() error {
 		log = cacheStore
 
 		// Create the snapshot store.
-		snapshots, err := raft.NewFileSnapshotStore(path, snapshotsRetained, s.config.LogOutput)
+		snapshots, err := raft.NewFileSnapshotStoreWithLogger(path, snapshotsRetained, s.logger.Named("snapshot"))
 		if err != nil {
 			return err
 		}
@@ -776,11 +833,33 @@ func (s *Server) setupRPC() error {
 	// been configured.
 	s.insecureRPCServer.Register(&AutoEncrypt{srv: s})
 
+	// Setup the AutoConfig JWT Authorizer
+	var authz AutoConfigAuthorizer
+	if s.config.AutoConfigAuthzEnabled {
+		// create the auto config authorizer from the JWT authmethod
+		validator, err := ssoauth.NewValidator(s.logger, &s.config.AutoConfigAuthzAuthMethod)
+		if err != nil {
+			return fmt.Errorf("Failed to initialize JWT Auto Config Authorizer: %w", err)
+		}
+
+		authz = &jwtAuthorizer{
+			validator:       validator,
+			allowReuse:      s.config.AutoConfigAuthzAllowReuse,
+			claimAssertions: s.config.AutoConfigAuthzClaimAssertions,
+		}
+	} else {
+		// This authorizer always returns that the endpoint is disabled
+		authz = &disabledAuthorizer{}
+	}
+	// now register with the insecure RPC server
+	s.insecureRPCServer.Register(NewAutoConfig(s.config, s.tlsConfigurator, s, authz))
+
 	ln, err := net.ListenTCP("tcp", s.config.RPCAddr)
 	if err != nil {
 		return err
 	}
 	s.Listener = ln
+
 	if s.config.NotifyListen != nil {
 		s.config.NotifyListen()
 	}
@@ -857,12 +936,16 @@ func (s *Server) Shutdown() error {
 	}
 
 	// Close the connection pool
-	s.connPool.Shutdown()
+	if s.connPool != nil {
+		s.connPool.Shutdown()
+	}
 
-	s.acls.Close()
+	if s.acls != nil {
+		s.acls.Close()
+	}
 
-	if s.config.NotifyShutdown != nil {
-		s.config.NotifyShutdown()
+	if s.fsm != nil {
+		s.fsm.State().Abandon()
 	}
 
 	return nil
@@ -1009,6 +1092,41 @@ func (s *Server) JoinWAN(addrs []string) (int, error) {
 	return s.serfWAN.Join(addrs, true)
 }
 
+// PrimaryMeshGatewayAddressesReadyCh returns a channel that will be closed
+// when federation state replication ships back at least one primary mesh
+// gateway (not via fallback config).
+func (s *Server) PrimaryMeshGatewayAddressesReadyCh() <-chan struct{} {
+	if s.gatewayLocator == nil {
+		return nil
+	}
+	return s.gatewayLocator.PrimaryMeshGatewayAddressesReadyCh()
+}
+
+// PickRandomMeshGatewaySuitableForDialing is a convenience function used for writing tests.
+func (s *Server) PickRandomMeshGatewaySuitableForDialing(dc string) string {
+	if s.gatewayLocator == nil {
+		return ""
+	}
+	return s.gatewayLocator.PickGateway(dc)
+}
+
+// RefreshPrimaryGatewayFallbackAddresses is used to update the list of current
+// fallback addresses for locating mesh gateways in the primary datacenter.
+func (s *Server) RefreshPrimaryGatewayFallbackAddresses(addrs []string) {
+	if s.gatewayLocator != nil {
+		s.gatewayLocator.RefreshPrimaryGatewayFallbackAddresses(addrs)
+	}
+}
+
+// PrimaryGatewayFallbackAddresses returns the current set of discovered
+// fallback addresses for the mesh gateways in the primary datacenter.
+func (s *Server) PrimaryGatewayFallbackAddresses() []string {
+	if s.gatewayLocator == nil {
+		return nil
+	}
+	return s.gatewayLocator.PrimaryGatewayFallbackAddresses()
+}
+
 // LocalMember is used to return the local node
 func (s *Server) LocalMember() serf.Member {
 	return s.serfLAN.LocalMember()
@@ -1057,6 +1175,12 @@ func (s *Server) IsLeader() bool {
 	return s.raft.State() == raft.Leader
 }
 
+// LeaderLastContact returns the time of last contact by a leader.
+// This only makes sense if we are currently a follower.
+func (s *Server) LeaderLastContact() time.Time {
+	return s.raft.LastContact()
+}
+
 // KeyManagerLAN returns the LAN Serf keyring manager
 func (s *Server) KeyManagerLAN() *serf.KeyManager {
 	return s.serfLAN.KeyManager()
@@ -1067,15 +1191,6 @@ func (s *Server) KeyManagerWAN() *serf.KeyManager {
 	return s.serfWAN.KeyManager()
 }
 
-// Encrypted determines if gossip is encrypted
-func (s *Server) Encrypted() bool {
-	LANEncrypted := s.serfLAN.EncryptionEnabled()
-	if s.serfWAN == nil {
-		return LANEncrypted
-	}
-	return LANEncrypted && s.serfWAN.EncryptionEnabled()
-}
-
 // LANSegments returns a map of LAN segments by name
 func (s *Server) LANSegments() map[string]*serf.Serf {
 	segments := make(map[string]*serf.Serf, len(s.segmentLAN)+1)
@@ -1301,6 +1416,72 @@ func (s *Server) intentionReplicationEnabled() bool {
 	return s.config.ConnectEnabled && s.config.Datacenter != s.config.PrimaryDatacenter
 }
 
+// CreateACLToken will create an ACL token from the given template
+func (s *Server) CreateACLToken(template *structs.ACLToken) (*structs.ACLToken, error) {
+	// we have to require local tokens or else it would require having these servers use a token with acl:write to make a
+	// token create RPC to the servers in the primary DC.
+	if !s.LocalTokensEnabled() {
+		return nil, fmt.Errorf("Agent Auto Configuration requires local token usage to be enabled in this datacenter: %s", s.config.Datacenter)
+	}
+
+	newToken := *template
+
+	// generate the accessor id
+	if newToken.AccessorID == "" {
+		accessor, err := lib.GenerateUUID(s.checkTokenUUID)
+		if err != nil {
+			return nil, err
+		}
+
+		newToken.AccessorID = accessor
+	}
+
+	// generate the secret id
+	if newToken.SecretID == "" {
+		secret, err := lib.GenerateUUID(s.checkTokenUUID)
+		if err != nil {
+			return nil, err
+		}
+
+		newToken.SecretID = secret
+	}
+
+	newToken.CreateTime = time.Now()
+
+	req := structs.ACLTokenBatchSetRequest{
+		Tokens: structs.ACLTokens{&newToken},
+		CAS:    false,
+	}
+
+	// perform the request to mint the new token
+	if _, err := s.raftApplyMsgpack(structs.ACLTokenSetRequestType, &req); err != nil {
+		return nil, err
+	}
+
+	// return the full token definition from the FSM
+	_, token, err := s.fsm.State().ACLTokenGetByAccessor(nil, newToken.AccessorID, &newToken.EnterpriseMeta)
+	return token, err
+}
+
+// DatacenterJoinAddresses will return all the strings suitable for usage in
+// retry join operations to connect to the the LAN or LAN segment gossip pool.
+func (s *Server) DatacenterJoinAddresses(segment string) ([]string, error) {
+	members, err := s.LANSegmentMembers(segment)
+	if err != nil {
+		return nil, fmt.Errorf("Failed to retrieve members for segment %s - %w", segment, err)
+	}
+
+	var joinAddrs []string
+	for _, m := range members {
+		if ok, _ := metadata.IsConsulServer(m); ok {
+			serfAddr := net.TCPAddr{IP: m.Addr, Port: int(m.Port)}
+			joinAddrs = append(joinAddrs, serfAddr.String())
+		}
+	}
+
+	return joinAddrs, nil
+}
+
 // peersInfoContent is used to help operators understand what happened to the
 // peers.json file. This is written to a file called peers.info in the same
 // location.
