@@ -2,15 +2,23 @@ package consul
 
 import (
 	"bytes"
+	"crypto/ecdsa"
+	"crypto/elliptic"
+	"crypto/rand"
+	"crypto/x509"
 	"fmt"
 	"net"
+	"net/rpc"
 	"os"
 	"strings"
 	"sync/atomic"
 	"testing"
 	"time"
 
+	"github.com/google/tcpproxy"
 	"github.com/hashicorp/consul/agent/connect/ca"
+	"github.com/hashicorp/consul/ipaddr"
+	"github.com/hashicorp/memberlist"
 
 	"github.com/hashicorp/consul/agent/connect"
 	"github.com/hashicorp/consul/agent/metadata"
@@ -26,6 +34,7 @@ import (
 	"github.com/hashicorp/go-uuid"
 	"golang.org/x/time/rate"
 
+	"github.com/stretchr/testify/assert"
 	"github.com/stretchr/testify/require"
 )
 
@@ -33,6 +42,46 @@ const (
 	TestDefaultMasterToken = "d9f05e83-a7ae-47ce-839e-c0d53a68c00a"
 )
 
+// testTLSCertificates Generates a TLS CA and server key/cert and returns them
+// in PEM encoded form.
+func testTLSCertificates(serverName string) (cert string, key string, cacert string, err error) {
+	// generate CA
+	serial, err := tlsutil.GenerateSerialNumber()
+	if err != nil {
+		return "", "", "", err
+	}
+	signer, err := ecdsa.GenerateKey(elliptic.P256(), rand.Reader)
+	if err != nil {
+		return "", "", "", err
+	}
+	ca, err := tlsutil.GenerateCA(signer, serial, 365, nil)
+	if err != nil {
+		return "", "", "", err
+	}
+
+	// generate leaf
+	serial, err = tlsutil.GenerateSerialNumber()
+	if err != nil {
+		return "", "", "", err
+	}
+
+	cert, privateKey, err := tlsutil.GenerateCert(
+		signer,
+		ca,
+		serial,
+		"Test Cert Name",
+		365,
+		[]string{serverName},
+		nil,
+		[]x509.ExtKeyUsage{x509.ExtKeyUsageServerAuth, x509.ExtKeyUsageClientAuth},
+	)
+	if err != nil {
+		return "", "", "", err
+	}
+
+	return cert, privateKey, ca, nil
+}
+
 // testServerACLConfig wraps another arbitrary Config altering callback
 // to setup some common ACL configurations. A new callback func will
 // be returned that has the original callback invoked after setting
@@ -43,7 +92,6 @@ func testServerACLConfig(cb func(*Config)) func(*Config) {
 		c.ACLsEnabled = true
 		c.ACLMasterToken = TestDefaultMasterToken
 		c.ACLDefaultPolicy = "deny"
-		c.ACLEnforceVersion8 = true
 
 		if cb != nil {
 			cb(c)
@@ -60,6 +108,7 @@ func configureTLS(config *Config) {
 var id int64
 
 func uniqueNodeName(name string) string {
+	name = strings.ReplaceAll(name, "/", "_")
 	return fmt.Sprintf("%s-node-%d", name, atomic.AddInt64(&id, 1))
 }
 
@@ -80,24 +129,20 @@ func waitForLeaderEstablishment(t *testing.T, servers ...*Server) {
 
 func testServerConfig(t *testing.T) (string, *Config) {
 	dir := testutil.TempDir(t, "consul")
+	t.Cleanup(func() {
+		os.RemoveAll(dir)
+	})
 	config := DefaultConfig()
 
 	ports := freeport.MustTake(3)
-
-	returnPortsFn := func() {
-		// The method of plumbing this into the server shutdown hook doesn't
-		// cover all exit points, so we insulate this against multiple
-		// invocations and then it's safe to call it a bunch of times.
+	t.Cleanup(func() {
 		freeport.Return(ports)
-		config.NotifyShutdown = nil // self-erasing
-	}
-	config.NotifyShutdown = returnPortsFn
+	})
 
 	config.NodeName = uniqueNodeName(t.Name())
 	config.Bootstrap = true
 	config.Datacenter = "dc1"
 	config.DataDir = dir
-	config.LogOutput = testutil.TestWriter(t)
 
 	// bind the rpc server to a random port. config.RPCAdvertise will be
 	// set to the listen address unless it was set in the configuration.
@@ -106,7 +151,6 @@ func testServerConfig(t *testing.T) (string, *Config) {
 
 	nodeID, err := uuid.GenerateUUID()
 	if err != nil {
-		returnPortsFn()
 		t.Fatal(err)
 	}
 	config.NodeID = types.NodeID(nodeID)
@@ -163,9 +207,6 @@ func testServerConfig(t *testing.T) (string, *Config) {
 			"IntermediateCertTTL": "288h",
 		},
 	}
-
-	config.NotifyShutdown = returnPortsFn
-
 	return dir, config
 }
 
@@ -209,21 +250,19 @@ func testServerDCExpectNonVoter(t *testing.T, dc string, expect int) (string, *S
 
 func testServerWithConfig(t *testing.T, cb func(*Config)) (string, *Server) {
 	var dir string
-	var config *Config
 	var srv *Server
-	var err error
 
 	// Retry added to avoid cases where bind addr is already in use
 	retry.RunWith(retry.ThreeTimes(), t, func(r *retry.R) {
+		var config *Config
 		dir, config = testServerConfig(t)
 		if cb != nil {
 			cb(config)
 		}
 
-		srv, err = newServer(config)
+		var err error
+		srv, err = newServer(t, config)
 		if err != nil {
-			config.NotifyShutdown()
-			os.RemoveAll(dir)
 			r.Fatalf("err: %v", err)
 		}
 	})
@@ -231,17 +270,21 @@ func testServerWithConfig(t *testing.T, cb func(*Config)) (string, *Server) {
 }
 
 // cb is a function that can alter the test servers configuration prior to the server starting.
-func testACLServerWithConfig(t *testing.T, cb func(*Config), initReplicationToken bool) (string, *Server) {
+func testACLServerWithConfig(t *testing.T, cb func(*Config), initReplicationToken bool) (string, *Server, rpc.ClientCodec) {
 	dir, srv := testServerWithConfig(t, testServerACLConfig(cb))
+	t.Cleanup(func() { srv.Shutdown() })
 
 	if initReplicationToken {
 		// setup some tokens here so we get less warnings in the logs
 		srv.tokens.UpdateReplicationToken(TestDefaultMasterToken, token.TokenSourceConfig)
 	}
-	return dir, srv
+
+	codec := rpcClient(t, srv)
+	t.Cleanup(func() { codec.Close() })
+	return dir, srv, codec
 }
 
-func newServer(c *Config) (*Server, error) {
+func newServer(t *testing.T, c *Config) (*Server, error) {
 	// chain server up notification
 	oldNotify := c.NotifyListen
 	up := make(chan struct{})
@@ -252,21 +295,19 @@ func newServer(c *Config) (*Server, error) {
 		}
 	}
 
-	// start server
-	w := c.LogOutput
-	if w == nil {
-		w = os.Stderr
-	}
 	logger := hclog.NewInterceptLogger(&hclog.LoggerOptions{
 		Name:   c.NodeName,
 		Level:  hclog.Debug,
-		Output: w,
+		Output: testutil.NewLogBuffer(t),
 	})
 	tlsConf, err := tlsutil.NewConfigurator(c.ToTLSUtilConfig(), logger)
 	if err != nil {
 		return nil, err
 	}
-	srv, err := NewServerLogger(c, logger, new(token.Store), tlsConf)
+	srv, err := NewServer(c,
+		WithLogger(logger),
+		WithTokenStore(new(token.Store)),
+		WithTLSConfigurator(tlsConf))
 	if err != nil {
 		return nil, err
 	}
@@ -291,8 +332,7 @@ func newServer(c *Config) (*Server, error) {
 func TestServer_StartStop(t *testing.T) {
 	t.Parallel()
 	// Start up a server and then stop it.
-	dir1, s1 := testServer(t)
-	defer os.RemoveAll(dir1)
+	_, s1 := testServer(t)
 	if err := s1.Shutdown(); err != nil {
 		t.Fatalf("err: %v", err)
 	}
@@ -306,20 +346,18 @@ func TestServer_StartStop(t *testing.T) {
 func TestServer_fixupACLDatacenter(t *testing.T) {
 	t.Parallel()
 
-	dir1, s1 := testServerWithConfig(t, func(c *Config) {
+	_, s1 := testServerWithConfig(t, func(c *Config) {
 		c.Datacenter = "aye"
 		c.PrimaryDatacenter = "aye"
 		c.ACLsEnabled = true
 	})
-	defer os.RemoveAll(dir1)
 	defer s1.Shutdown()
 
-	dir2, s2 := testServerWithConfig(t, func(c *Config) {
+	_, s2 := testServerWithConfig(t, func(c *Config) {
 		c.Datacenter = "bee"
 		c.PrimaryDatacenter = "aye"
 		c.ACLsEnabled = true
 	})
-	defer os.RemoveAll(dir2)
 	defer s2.Shutdown()
 
 	// Try to join
@@ -367,6 +405,68 @@ func TestServer_JoinLAN(t *testing.T) {
 	})
 }
 
+// TestServer_JoinLAN_SerfAllowedCIDRs test that IPs might be blocked
+// with Serf.
+// To run properly, this test requires to be able to bind and have access
+// on 127.0.1.1 which is the case for most Linux machines and Windows,
+// so Unit test will run in the CI.
+// To run it on Mac OS, please run this commandd first, otherwise the
+// test will be skipped: `sudo ifconfig lo0 alias 127.0.1.1 up`
+func TestServer_JoinLAN_SerfAllowedCIDRs(t *testing.T) {
+	t.Parallel()
+	dir1, s1 := testServerWithConfig(t, func(c *Config) {
+		c.BootstrapExpect = 1
+		lan, err := memberlist.ParseCIDRs([]string{"127.0.0.1/32"})
+		assert.NoError(t, err)
+		c.SerfLANConfig.MemberlistConfig.CIDRsAllowed = lan
+		wan, err := memberlist.ParseCIDRs([]string{"127.0.0.0/24", "::1/128"})
+		assert.NoError(t, err)
+		c.SerfWANConfig.MemberlistConfig.CIDRsAllowed = wan
+	})
+	defer os.RemoveAll(dir1)
+	defer s1.Shutdown()
+
+	targetAddr := "127.0.1.1"
+	dir2, a2, err := testClientWithConfigWithErr(t, func(c *Config) {
+		c.SerfLANConfig.MemberlistConfig.BindAddr = targetAddr
+	})
+	defer os.RemoveAll(dir2)
+	if err != nil {
+		t.Skipf("Cannot bind on %s, to run on Mac OS: `sudo ifconfig lo0 alias 127.0.1.1 up`", targetAddr)
+	}
+	defer a2.Shutdown()
+
+	dir3, rs3 := testServerWithConfig(t, func(c *Config) {
+		c.BootstrapExpect = 1
+		c.Datacenter = "dc2"
+	})
+	defer os.RemoveAll(dir3)
+	defer rs3.Shutdown()
+
+	leaderAddr := joinAddrLAN(s1)
+	if _, err := a2.JoinLAN([]string{leaderAddr}); err != nil {
+		t.Fatalf("Expected no error, had: %#v", err)
+	}
+	// Try to join
+	joinWAN(t, rs3, s1)
+	retry.Run(t, func(r *retry.R) {
+		if got, want := len(s1.LANMembers()), 1; got != want {
+			// LAN is blocked, should be 1 only
+			r.Fatalf("got %d s1 LAN members want %d", got, want)
+		}
+		if got, want := len(a2.LANMembers()), 2; got != want {
+			// LAN is blocked a2 can see s1, but not s1
+			r.Fatalf("got %d a2 LAN members want %d", got, want)
+		}
+		if got, want := len(s1.WANMembers()), 2; got != want {
+			r.Fatalf("got %d s1 WAN members want %d", got, want)
+		}
+		if got, want := len(rs3.WANMembers()), 2; got != want {
+			r.Fatalf("got %d rs3 WAN members want %d", got, want)
+		}
+	})
+}
+
 func TestServer_LANReap(t *testing.T) {
 	t.Parallel()
 
@@ -543,6 +643,239 @@ func TestServer_JoinWAN_Flood(t *testing.T) {
 	}
 }
 
+// This is a mirror of a similar test in agent/agent_test.go
+func TestServer_JoinWAN_viaMeshGateway(t *testing.T) {
+	t.Parallel()
+
+	gwPort := freeport.MustTake(1)
+	defer freeport.Return(gwPort)
+	gwAddr := ipaddr.FormatAddressPort("127.0.0.1", gwPort[0])
+
+	dir1, s1 := testServerWithConfig(t, func(c *Config) {
+		c.Domain = "consul"
+		c.NodeName = "bob"
+		c.Datacenter = "dc1"
+		c.PrimaryDatacenter = "dc1"
+		c.Bootstrap = true
+		// tls
+		c.CAFile = "../../test/hostname/CertAuth.crt"
+		c.CertFile = "../../test/hostname/Bob.crt"
+		c.KeyFile = "../../test/hostname/Bob.key"
+		c.VerifyIncoming = true
+		c.VerifyOutgoing = true
+		c.VerifyServerHostname = true
+		// wanfed
+		c.ConnectMeshGatewayWANFederationEnabled = true
+	})
+	defer os.RemoveAll(dir1)
+	defer s1.Shutdown()
+
+	dir2, s2 := testServerWithConfig(t, func(c *Config) {
+		c.Domain = "consul"
+		c.NodeName = "betty"
+		c.Datacenter = "dc2"
+		c.PrimaryDatacenter = "dc1"
+		c.Bootstrap = true
+		// tls
+		c.CAFile = "../../test/hostname/CertAuth.crt"
+		c.CertFile = "../../test/hostname/Betty.crt"
+		c.KeyFile = "../../test/hostname/Betty.key"
+		c.VerifyIncoming = true
+		c.VerifyOutgoing = true
+		c.VerifyServerHostname = true
+		// wanfed
+		c.ConnectMeshGatewayWANFederationEnabled = true
+	})
+	defer os.RemoveAll(dir2)
+	defer s2.Shutdown()
+
+	dir3, s3 := testServerWithConfig(t, func(c *Config) {
+		c.Domain = "consul"
+		c.NodeName = "bonnie"
+		c.Datacenter = "dc3"
+		c.PrimaryDatacenter = "dc1"
+		c.Bootstrap = true
+		// tls
+		c.CAFile = "../../test/hostname/CertAuth.crt"
+		c.CertFile = "../../test/hostname/Bonnie.crt"
+		c.KeyFile = "../../test/hostname/Bonnie.key"
+		c.VerifyIncoming = true
+		c.VerifyOutgoing = true
+		c.VerifyServerHostname = true
+		// wanfed
+		c.ConnectMeshGatewayWANFederationEnabled = true
+	})
+	defer os.RemoveAll(dir3)
+	defer s3.Shutdown()
+
+	// We'll use the same gateway for all datacenters since it doesn't care.
+	var p tcpproxy.Proxy
+	p.AddSNIRoute(gwAddr, "bob.server.dc1.consul", tcpproxy.To(s1.config.RPCAddr.String()))
+	p.AddSNIRoute(gwAddr, "betty.server.dc2.consul", tcpproxy.To(s2.config.RPCAddr.String()))
+	p.AddSNIRoute(gwAddr, "bonnie.server.dc3.consul", tcpproxy.To(s3.config.RPCAddr.String()))
+	p.AddStopACMESearch(gwAddr)
+	require.NoError(t, p.Start())
+	defer func() {
+		p.Close()
+		p.Wait()
+	}()
+
+	t.Logf("routing %s => %s", "bob.server.dc1.consul", s1.config.RPCAddr.String())
+	t.Logf("routing %s => %s", "betty.server.dc2.consul", s2.config.RPCAddr.String())
+	t.Logf("routing %s => %s", "bonnie.server.dc3.consul", s3.config.RPCAddr.String())
+
+	// Register this into the catalog in dc1.
+	{
+		arg := structs.RegisterRequest{
+			Datacenter: "dc1",
+			Node:       "bob",
+			Address:    "127.0.0.1",
+			Service: &structs.NodeService{
+				Kind:    structs.ServiceKindMeshGateway,
+				ID:      "mesh-gateway",
+				Service: "mesh-gateway",
+				Meta:    map[string]string{structs.MetaWANFederationKey: "1"},
+				Port:    gwPort[0],
+			},
+		}
+
+		var out struct{}
+		require.NoError(t, s1.RPC("Catalog.Register", &arg, &out))
+	}
+
+	// Wait for it to make it into the gateway locator.
+	retry.Run(t, func(r *retry.R) {
+		require.NotEmpty(r, s1.gatewayLocator.PickGateway("dc1"))
+	})
+
+	// Seed the secondaries with the address of the primary and wait for that to
+	// be in their locators.
+	s2.RefreshPrimaryGatewayFallbackAddresses([]string{gwAddr})
+	retry.Run(t, func(r *retry.R) {
+		require.NotEmpty(r, s2.gatewayLocator.PickGateway("dc1"))
+	})
+	s3.RefreshPrimaryGatewayFallbackAddresses([]string{gwAddr})
+	retry.Run(t, func(r *retry.R) {
+		require.NotEmpty(r, s3.gatewayLocator.PickGateway("dc1"))
+	})
+
+	// Try to join from secondary to primary. We can't use joinWAN() because we
+	// are simulating proper bootstrapping and if ACLs were on we would have to
+	// delay gateway registration in the secondary until after one directional
+	// join. So this way we explicitly join secondary-to-primary as a standalone
+	// operation and follow it up later with a full join.
+	_, err := s2.JoinWAN([]string{joinAddrWAN(s1)})
+	require.NoError(t, err)
+	retry.Run(t, func(r *retry.R) {
+		if got, want := len(s2.WANMembers()), 2; got != want {
+			r.Fatalf("got %d s2 WAN members want %d", got, want)
+		}
+	})
+	_, err = s3.JoinWAN([]string{joinAddrWAN(s1)})
+	require.NoError(t, err)
+	retry.Run(t, func(r *retry.R) {
+		if got, want := len(s3.WANMembers()), 3; got != want {
+			r.Fatalf("got %d s3 WAN members want %d", got, want)
+		}
+	})
+
+	// Now we can register this into the catalog in dc2 and dc3.
+	{
+		arg := structs.RegisterRequest{
+			Datacenter: "dc2",
+			Node:       "betty",
+			Address:    "127.0.0.1",
+			Service: &structs.NodeService{
+				Kind:    structs.ServiceKindMeshGateway,
+				ID:      "mesh-gateway",
+				Service: "mesh-gateway",
+				Meta:    map[string]string{structs.MetaWANFederationKey: "1"},
+				Port:    gwPort[0],
+			},
+		}
+
+		var out struct{}
+		require.NoError(t, s2.RPC("Catalog.Register", &arg, &out))
+	}
+	{
+		arg := structs.RegisterRequest{
+			Datacenter: "dc3",
+			Node:       "bonnie",
+			Address:    "127.0.0.1",
+			Service: &structs.NodeService{
+				Kind:    structs.ServiceKindMeshGateway,
+				ID:      "mesh-gateway",
+				Service: "mesh-gateway",
+				Meta:    map[string]string{structs.MetaWANFederationKey: "1"},
+				Port:    gwPort[0],
+			},
+		}
+
+		var out struct{}
+		require.NoError(t, s3.RPC("Catalog.Register", &arg, &out))
+	}
+
+	// Wait for it to make it into the gateway locator in dc2 and then for
+	// AE to carry it back to the primary
+	retry.Run(t, func(r *retry.R) {
+		require.NotEmpty(r, s3.gatewayLocator.PickGateway("dc2"))
+		require.NotEmpty(r, s2.gatewayLocator.PickGateway("dc2"))
+		require.NotEmpty(r, s1.gatewayLocator.PickGateway("dc2"))
+
+		require.NotEmpty(r, s3.gatewayLocator.PickGateway("dc3"))
+		require.NotEmpty(r, s2.gatewayLocator.PickGateway("dc3"))
+		require.NotEmpty(r, s1.gatewayLocator.PickGateway("dc3"))
+	})
+
+	// Try to join again using the standard verification method now that
+	// all of the plumbing is in place.
+	joinWAN(t, s2, s1)
+	retry.Run(t, func(r *retry.R) {
+		if got, want := len(s1.WANMembers()), 3; got != want {
+			r.Fatalf("got %d s1 WAN members want %d", got, want)
+		}
+		if got, want := len(s2.WANMembers()), 3; got != want {
+			r.Fatalf("got %d s2 WAN members want %d", got, want)
+		}
+	})
+
+	// Check the router has all of them
+	retry.Run(t, func(r *retry.R) {
+		if got, want := len(s1.router.GetDatacenters()), 3; got != want {
+			r.Fatalf("got %d routes want %d", got, want)
+		}
+		if got, want := len(s2.router.GetDatacenters()), 3; got != want {
+			r.Fatalf("got %d datacenters want %d", got, want)
+		}
+		if got, want := len(s3.router.GetDatacenters()), 3; got != want {
+			r.Fatalf("got %d datacenters want %d", got, want)
+		}
+	})
+
+	// Ensure we can do some trivial RPC in all directions.
+	servers := map[string]*Server{"dc1": s1, "dc2": s2, "dc3": s3}
+	names := map[string]string{"dc1": "bob", "dc2": "betty", "dc3": "bonnie"}
+	for _, srcDC := range []string{"dc1", "dc2", "dc3"} {
+		srv := servers[srcDC]
+		for _, dstDC := range []string{"dc1", "dc2", "dc3"} {
+			if srcDC == dstDC {
+				continue
+			}
+			t.Run(srcDC+" to "+dstDC, func(t *testing.T) {
+				arg := structs.DCSpecificRequest{
+					Datacenter: dstDC,
+				}
+				var out structs.IndexedNodes
+				require.NoError(t, srv.RPC("Catalog.ListNodes", &arg, &out))
+				require.Len(t, out.Nodes, 1)
+				node := out.Nodes[0]
+				require.Equal(t, dstDC, node.Datacenter)
+				require.Equal(t, names[dstDC], node.Node)
+			})
+		}
+	}
+}
+
 func TestServer_JoinSeparateLanAndWanAddresses(t *testing.T) {
 	t.Parallel()
 	dir1, s1 := testServerWithConfig(t, func(c *Config) {
@@ -735,28 +1068,26 @@ func TestServer_RPC(t *testing.T) {
 
 func TestServer_JoinLAN_TLS(t *testing.T) {
 	t.Parallel()
-	dir1, conf1 := testServerConfig(t)
+	_, conf1 := testServerConfig(t)
 	conf1.VerifyIncoming = true
 	conf1.VerifyOutgoing = true
 	configureTLS(conf1)
-	s1, err := newServer(conf1)
+	s1, err := newServer(t, conf1)
 	if err != nil {
 		t.Fatalf("err: %v", err)
 	}
-	defer os.RemoveAll(dir1)
 	defer s1.Shutdown()
 	testrpc.WaitForTestAgent(t, s1.RPC, "dc1")
 
-	dir2, conf2 := testServerConfig(t)
+	_, conf2 := testServerConfig(t)
 	conf2.Bootstrap = false
 	conf2.VerifyIncoming = true
 	conf2.VerifyOutgoing = true
 	configureTLS(conf2)
-	s2, err := newServer(conf2)
+	s2, err := newServer(t, conf2)
 	if err != nil {
 		t.Fatalf("err: %v", err)
 	}
-	defer os.RemoveAll(dir2)
 	defer s2.Shutdown()
 
 	// Try to join
@@ -940,7 +1271,7 @@ func (r *fakeGlobalResp) New() interface{} {
 	return struct{}{}
 }
 
-func TestServer_globalRPCErrors(t *testing.T) {
+func TestServer_keyringRPCs(t *testing.T) {
 	t.Parallel()
 	dir1, s1 := testServerDC(t, "dc1")
 	defer os.RemoveAll(dir1)
@@ -952,7 +1283,7 @@ func TestServer_globalRPCErrors(t *testing.T) {
 	})
 
 	// Check that an error from a remote DC is returned
-	err := s1.globalRPC("Bad.Method", nil, &fakeGlobalResp{})
+	_, err := s1.keyringRPCs("Bad.Method", nil, []string{s1.config.Datacenter})
 	if err == nil {
 		t.Fatalf("should have errored")
 	}
@@ -961,28 +1292,6 @@ func TestServer_globalRPCErrors(t *testing.T) {
 	}
 }
 
-func TestServer_Encrypted(t *testing.T) {
-	t.Parallel()
-	dir1, s1 := testServer(t)
-	defer os.RemoveAll(dir1)
-	defer s1.Shutdown()
-
-	key := []byte{0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15}
-	dir2, s2 := testServerWithConfig(t, func(c *Config) {
-		c.SerfLANConfig.MemberlistConfig.SecretKey = key
-		c.SerfWANConfig.MemberlistConfig.SecretKey = key
-	})
-	defer os.RemoveAll(dir2)
-	defer s2.Shutdown()
-
-	if s1.Encrypted() {
-		t.Fatalf("should not be encrypted")
-	}
-	if !s2.Encrypted() {
-		t.Fatalf("should be encrypted")
-	}
-}
-
 func testVerifyRPC(s1, s2 *Server, t *testing.T) (bool, error) {
 	joinLAN(t, s1, s2)
 	retry.Run(t, func(r *retry.R) {
@@ -999,7 +1308,7 @@ func testVerifyRPC(s1, s2 *Server, t *testing.T) (bool, error) {
 	if leader == nil {
 		t.Fatal("no leader")
 	}
-	return s2.connPool.Ping(leader.Datacenter, leader.Addr, leader.Version)
+	return s2.connPool.Ping(leader.Datacenter, leader.ShortName, leader.Addr)
 }
 
 func TestServer_TLSToNoTLS(t *testing.T) {
@@ -1156,14 +1465,13 @@ func TestServer_Reload(t *testing.T) {
 
 func TestServer_RPC_RateLimit(t *testing.T) {
 	t.Parallel()
-	dir1, conf1 := testServerConfig(t)
+	_, conf1 := testServerConfig(t)
 	conf1.RPCRate = 2
 	conf1.RPCMaxBurst = 2
-	s1, err := NewServer(conf1)
+	s1, err := newServer(t, conf1)
 	if err != nil {
 		t.Fatalf("err: %v", err)
 	}
-	defer os.RemoveAll(dir1)
 	defer s1.Shutdown()
 	testrpc.WaitForLeader(t, s1.RPC, "dc1")
 
@@ -1177,7 +1485,7 @@ func TestServer_RPC_RateLimit(t *testing.T) {
 
 func TestServer_CALogging(t *testing.T) {
 	t.Parallel()
-	dir1, conf1 := testServerConfig(t)
+	_, conf1 := testServerConfig(t)
 
 	// Setup dummy logger to catch output
 	var buf bytes.Buffer
@@ -1185,11 +1493,14 @@ func TestServer_CALogging(t *testing.T) {
 
 	c, err := tlsutil.NewConfigurator(conf1.ToTLSUtilConfig(), logger)
 	require.NoError(t, err)
-	s1, err := NewServerLogger(conf1, logger, new(token.Store), c)
+
+	s1, err := NewServer(conf1,
+		WithLogger(logger),
+		WithTokenStore(new(token.Store)),
+		WithTLSConfigurator(c))
 	if err != nil {
 		t.Fatalf("err: %v", err)
 	}
-	defer os.RemoveAll(dir1)
 	defer s1.Shutdown()
 	testrpc.WaitForLeader(t, s1.RPC, "dc1")
 
@@ -1207,3 +1518,97 @@ func TestServer_CALogging(t *testing.T) {
 
 	require.Contains(t, buf.String(), "consul CA provider configured")
 }
+
+func TestServer_DatacenterJoinAddresses(t *testing.T) {
+	conf := testClusterConfig{
+		Datacenter: "primary",
+		Servers:    3,
+	}
+
+	nodes := newTestCluster(t, &conf)
+
+	var expected []string
+	for _, srv := range nodes.Servers {
+		expected = append(expected, fmt.Sprintf("127.0.0.1:%d", srv.config.SerfLANConfig.MemberlistConfig.BindPort))
+	}
+
+	actual, err := nodes.Servers[0].DatacenterJoinAddresses("")
+	require.NoError(t, err)
+	require.ElementsMatch(t, expected, actual)
+}
+
+func TestServer_CreateACLToken(t *testing.T) {
+	_, srv, codec := testACLServerWithConfig(t, nil, false)
+
+	waitForLeaderEstablishment(t, srv)
+
+	r1, err := upsertTestRole(codec, TestDefaultMasterToken, "dc1")
+	require.NoError(t, err)
+
+	t.Run("predefined-ids", func(t *testing.T) {
+		accessor := "554cd3ab-5d4e-4d6e-952e-4e8b6c77bfb3"
+		secret := "ef453f31-ad58-4ec8-8bf8-342e99763026"
+		in := &structs.ACLToken{
+			AccessorID:  accessor,
+			SecretID:    secret,
+			Description: "test",
+			Policies: []structs.ACLTokenPolicyLink{
+				{
+					ID: structs.ACLPolicyGlobalManagementID,
+				},
+			},
+			NodeIdentities: []*structs.ACLNodeIdentity{
+				{
+					NodeName:   "foo",
+					Datacenter: "bar",
+				},
+			},
+			ServiceIdentities: []*structs.ACLServiceIdentity{
+				{
+					ServiceName: "web",
+				},
+			},
+			Roles: []structs.ACLTokenRoleLink{
+				{
+					ID: r1.ID,
+				},
+			},
+		}
+
+		out, err := srv.CreateACLToken(in)
+		require.NoError(t, err)
+		require.Equal(t, accessor, out.AccessorID)
+		require.Equal(t, secret, out.SecretID)
+		require.Equal(t, "test", out.Description)
+		require.NotZero(t, out.CreateTime)
+		require.Len(t, out.Policies, 1)
+		require.Len(t, out.Roles, 1)
+		require.Len(t, out.NodeIdentities, 1)
+		require.Len(t, out.ServiceIdentities, 1)
+		require.Equal(t, structs.ACLPolicyGlobalManagementID, out.Policies[0].ID)
+		require.Equal(t, "foo", out.NodeIdentities[0].NodeName)
+		require.Equal(t, "web", out.ServiceIdentities[0].ServiceName)
+		require.Equal(t, r1.ID, out.Roles[0].ID)
+	})
+
+	t.Run("autogen-ids", func(t *testing.T) {
+		in := &structs.ACLToken{
+			Description: "test",
+			NodeIdentities: []*structs.ACLNodeIdentity{
+				{
+					NodeName:   "foo",
+					Datacenter: "bar",
+				},
+			},
+		}
+
+		out, err := srv.CreateACLToken(in)
+		require.NoError(t, err)
+		require.NotEmpty(t, out.AccessorID)
+		require.NotEmpty(t, out.SecretID)
+		require.Equal(t, "test", out.Description)
+		require.NotZero(t, out.CreateTime)
+		require.Len(t, out.NodeIdentities, 1)
+		require.Equal(t, "foo", out.NodeIdentities[0].NodeName)
+	})
+}
