@@ -5,8 +5,8 @@ import (
 	"fmt"
 	"time"
 
-	"github.com/hashicorp/consul/agent/agentpb"
 	"github.com/hashicorp/consul/agent/structs"
+	pbacl "github.com/hashicorp/consul/proto/pbacl"
 	memdb "github.com/hashicorp/go-memdb"
 )
 
@@ -340,7 +340,7 @@ func (s *Store) CanBootstrapACLToken() (bool, uint64, error) {
 // to update the name. Unlike the older functions to operate specifically on role or policy links
 // this function does not itself handle the case where the id cannot be found. Instead the
 // getName function should handle that and return an error if necessary
-func (s *Store) resolveACLLinks(tx *memdb.Txn, links []agentpb.ACLLink, getName func(*memdb.Txn, string) (string, error)) (int, error) {
+func (s *Store) resolveACLLinks(tx *memdb.Txn, links []pbacl.ACLLink, getName func(*memdb.Txn, string) (string, error)) (int, error) {
 	var numValid int
 	for linkIndex, link := range links {
 		if link.ID != "" {
@@ -366,12 +366,12 @@ func (s *Store) resolveACLLinks(tx *memdb.Txn, links []agentpb.ACLLink, getName
 // associated with the ID of the link. Ideally this will be a no-op if the names are already correct
 // however if a linked resource was renamed it might be stale. This function will treat the incoming
 // links with copy-on-write semantics and its output will indicate whether any modifications were made.
-func (s *Store) fixupACLLinks(tx *memdb.Txn, original []agentpb.ACLLink, getName func(*memdb.Txn, string) (string, error)) ([]agentpb.ACLLink, bool, error) {
+func (s *Store) fixupACLLinks(tx *memdb.Txn, original []pbacl.ACLLink, getName func(*memdb.Txn, string) (string, error)) ([]pbacl.ACLLink, bool, error) {
 	owned := false
 	links := original
 
-	cloneLinks := func(l []agentpb.ACLLink, copyNumLinks int) []agentpb.ACLLink {
-		clone := make([]agentpb.ACLLink, copyNumLinks)
+	cloneLinks := func(l []pbacl.ACLLink, copyNumLinks int) []pbacl.ACLLink {
+		clone := make([]pbacl.ACLLink, copyNumLinks)
 		copy(clone, l[:copyNumLinks])
 		return clone
 	}
@@ -397,7 +397,7 @@ func (s *Store) fixupACLLinks(tx *memdb.Txn, original []agentpb.ACLLink, getName
 			}
 
 			// append the corrected link
-			links = append(links, agentpb.ACLLink{ID: link.ID, Name: name})
+			links = append(links, pbacl.ACLLink{ID: link.ID, Name: name})
 		} else if owned {
 			links = append(links, link)
 		}
@@ -742,8 +742,17 @@ func (s *Store) aclTokenSetTxn(tx *memdb.Txn, idx uint64, token *structs.ACLToke
 		}
 	}
 
+	for _, nodeid := range token.NodeIdentities {
+		if nodeid.NodeName == "" {
+			return fmt.Errorf("Encountered a Token with an empty node identity name in the state store")
+		}
+		if nodeid.Datacenter == "" {
+			return fmt.Errorf("Encountered a Token with an empty node identity datacenter in the state store")
+		}
+	}
+
 	if prohibitUnprivileged {
-		if numValidRoles == 0 && numValidPolicies == 0 && len(token.ServiceIdentities) == 0 {
+		if numValidRoles == 0 && numValidPolicies == 0 && len(token.ServiceIdentities) == 0 && len(token.NodeIdentities) == 0 {
 			return ErrTokenHasNoPrivileges
 		}
 	}
@@ -1369,6 +1378,15 @@ func (s *Store) aclRoleSetTxn(tx *memdb.Txn, idx uint64, role *structs.ACLRole,
 		}
 	}
 
+	for _, nodeid := range role.NodeIdentities {
+		if nodeid.NodeName == "" {
+			return fmt.Errorf("Encountered a Role with an empty node identity name in the state store")
+		}
+		if nodeid.Datacenter == "" {
+			return fmt.Errorf("Encountered a Role with an empty node identity datacenter in the state store")
+		}
+	}
+
 	if err := s.aclRoleUpsertValidateEnterprise(tx, role, existing); err != nil {
 		return err
 	}
