@@ -4,7 +4,7 @@ import (
 	"container/list"
 	"crypto/tls"
 	"fmt"
-	"io"
+	"log"
 	"net"
 	"net/rpc"
 	"sync"
@@ -42,10 +42,10 @@ type Conn struct {
 	refCount    int32
 	shouldClose int32
 
+	nodeName string
 	addr     net.Addr
 	session  muxSession
 	lastUsed time.Time
-	version  int
 
 	pool *ConnPool
 
@@ -126,8 +126,9 @@ type ConnPool struct {
 	// SrcAddr is the source address for outgoing connections.
 	SrcAddr *net.TCPAddr
 
-	// LogOutput is used to control logging
-	LogOutput io.Writer
+	// Logger passed to yamux
+	// TODO: consider refactoring to accept a full yamux.Config instead of a logger
+	Logger *log.Logger
 
 	// The maximum time to keep a connection open
 	MaxTime time.Duration
@@ -138,12 +139,21 @@ type ConnPool struct {
 	// TLSConfigurator
 	TLSConfigurator *tlsutil.Configurator
 
-	// ForceTLS is used to enforce outgoing TLS verification
-	ForceTLS bool
+	// GatewayResolver is a function that returns a suitable random mesh
+	// gateway address for dialing servers in a given DC. This is only
+	// needed if wan federation via mesh gateways is enabled.
+	GatewayResolver func(string) string
+
+	// Datacenter is the datacenter of the current agent.
+	Datacenter string
+
+	// Server should be set to true if this connection pool is configured in a
+	// server instead of a client.
+	Server bool
 
 	sync.Mutex
 
-	// pool maps an address to a open connection
+	// pool maps a nodeName+address to a open connection
 	pool map[string]*Conn
 
 	// limiter is used to throttle the number of connect attempts
@@ -196,14 +206,20 @@ func (p *ConnPool) Shutdown() error {
 // wait for an existing connection attempt to finish, if one if in progress,
 // and will return that one if it succeeds. If all else fails, it will return a
 // newly-created connection and add it to the pool.
-func (p *ConnPool) acquire(dc string, addr net.Addr, version int, useTLS bool) (*Conn, error) {
+func (p *ConnPool) acquire(dc string, nodeName string, addr net.Addr) (*Conn, error) {
+	if nodeName == "" {
+		return nil, fmt.Errorf("pool: ConnPool.acquire requires a node name")
+	}
+
 	addrStr := addr.String()
 
+	poolKey := nodeName + ":" + addrStr
+
 	// Check to see if there's a pooled connection available. This is up
 	// here since it should the vastly more common case than the rest
 	// of the code here.
 	p.Lock()
-	c := p.pool[addrStr]
+	c := p.pool[poolKey]
 	if c != nil {
 		c.markForUse()
 		p.Unlock()
@@ -225,7 +241,7 @@ func (p *ConnPool) acquire(dc string, addr net.Addr, version int, useTLS bool) (
 	// If we are the lead thread, make the new connection and then wake
 	// everybody else up to see if we got it.
 	if isLeadThread {
-		c, err := p.getNewConn(dc, addr, version, useTLS)
+		c, err := p.getNewConn(dc, nodeName, addr)
 		p.Lock()
 		delete(p.limiter, addrStr)
 		close(wait)
@@ -234,7 +250,7 @@ func (p *ConnPool) acquire(dc string, addr net.Addr, version int, useTLS bool) (
 			return nil, err
 		}
 
-		p.pool[addrStr] = c
+		p.pool[poolKey] = c
 		p.Unlock()
 		return c, nil
 	}
@@ -249,7 +265,7 @@ func (p *ConnPool) acquire(dc string, addr net.Addr, version int, useTLS bool) (
 
 	// See if the lead thread was able to get us a connection.
 	p.Lock()
-	if c := p.pool[addrStr]; c != nil {
+	if c := p.pool[poolKey]; c != nil {
 		c.markForUse()
 		p.Unlock()
 		return c, nil
@@ -267,29 +283,50 @@ type HalfCloser interface {
 
 // DialTimeout is used to establish a raw connection to the given server, with
 // given connection timeout. It also writes RPCTLS as the first byte.
-func (p *ConnPool) DialTimeout(dc string, addr net.Addr, timeout time.Duration, useTLS bool) (net.Conn, HalfCloser, error) {
+func (p *ConnPool) DialTimeout(
+	dc string,
+	nodeName string,
+	addr net.Addr,
+	actualRPCType RPCType,
+) (net.Conn, HalfCloser, error) {
 	p.once.Do(p.init)
 
-	return DialTimeoutWithRPCType(dc, addr, p.SrcAddr, timeout, useTLS || p.ForceTLS, p.TLSConfigurator.OutgoingRPCWrapper(), RPCTLS)
-}
-
-// DialTimeoutInsecure is used to establish a raw connection to the given
-// server, with given connection timeout. It also writes RPCTLSInsecure as the
-// first byte to indicate that the client cannot provide a certificate. This is
-// so far only used for AutoEncrypt.Sign.
-func (p *ConnPool) DialTimeoutInsecure(dc string, addr net.Addr, timeout time.Duration, wrapper tlsutil.DCWrapper) (net.Conn, HalfCloser, error) {
-	p.once.Do(p.init)
-
-	if wrapper == nil {
-		return nil, nil, fmt.Errorf("wrapper cannot be nil")
+	if p.Server && p.GatewayResolver != nil && p.TLSConfigurator != nil && dc != p.Datacenter {
+		// NOTE: TLS is required on this branch.
+		return DialTimeoutWithRPCTypeViaMeshGateway(
+			dc,
+			nodeName,
+			addr,
+			p.SrcAddr,
+			p.TLSConfigurator.OutgoingALPNRPCWrapper(),
+			actualRPCType,
+			RPCTLS,
+			// gateway stuff
+			p.Server,
+			p.TLSConfigurator,
+			p.GatewayResolver,
+			p.Datacenter,
+		)
 	}
 
-	return DialTimeoutWithRPCType(dc, addr, p.SrcAddr, timeout, true, wrapper, RPCTLSInsecure)
+	return p.dial(
+		dc,
+		nodeName,
+		addr,
+		actualRPCType,
+		RPCTLS,
+	)
 }
 
-func DialTimeoutWithRPCType(dc string, addr net.Addr, src *net.TCPAddr, timeout time.Duration, useTLS bool, wrapper tlsutil.DCWrapper, rpcType RPCType) (net.Conn, HalfCloser, error) {
+func (p *ConnPool) dial(
+	dc string,
+	nodeName string,
+	addr net.Addr,
+	actualRPCType RPCType,
+	tlsRPCType RPCType,
+) (net.Conn, HalfCloser, error) {
 	// Try to dial the conn
-	d := &net.Dialer{LocalAddr: src, Timeout: timeout}
+	d := &net.Dialer{LocalAddr: p.SrcAddr, Timeout: defaultDialTimeout}
 	conn, err := d.Dial("tcp", addr.String())
 	if err != nil {
 		return nil, nil, err
@@ -306,9 +343,10 @@ func DialTimeoutWithRPCType(dc string, addr net.Addr, src *net.TCPAddr, timeout
 	}
 
 	// Check if TLS is enabled
-	if useTLS && wrapper != nil {
+	if p.TLSConfigurator.UseTLS(dc) {
+		wrapper := p.TLSConfigurator.OutgoingRPCWrapper()
 		// Switch the connection into TLS mode
-		if _, err := conn.Write([]byte{byte(rpcType)}); err != nil {
+		if _, err := conn.Write([]byte{byte(tlsRPCType)}); err != nil {
 			conn.Close()
 			return nil, nil, err
 		}
@@ -327,45 +365,123 @@ func DialTimeoutWithRPCType(dc string, addr net.Addr, src *net.TCPAddr, timeout
 		}
 	}
 
+	// Send the type-byte for the protocol if one is required.
+	//
+	// When using insecure TLS there is no inner type-byte as these connections
+	// aren't wrapped like the standard TLS ones are.
+	if tlsRPCType != RPCTLSInsecure {
+		if _, err := conn.Write([]byte{byte(actualRPCType)}); err != nil {
+			conn.Close()
+			return nil, nil, err
+		}
+	}
+
+	return conn, hc, nil
+}
+
+// DialTimeoutWithRPCTypeViaMeshGateway dials the destination node and sets up
+// the connection to be the correct RPC type using ALPN. This currently is
+// exclusively used to dial other servers in foreign datacenters via mesh
+// gateways.
+//
+// NOTE: There is a close mirror of this method in agent/consul/wanfed/wanfed.go:dial
+func DialTimeoutWithRPCTypeViaMeshGateway(
+	dc string,
+	nodeName string,
+	addr net.Addr,
+	src *net.TCPAddr,
+	wrapper tlsutil.ALPNWrapper,
+	actualRPCType RPCType,
+	tlsRPCType RPCType,
+	// gateway stuff
+	dialingFromServer bool,
+	tlsConfigurator *tlsutil.Configurator,
+	gatewayResolver func(string) string,
+	thisDatacenter string,
+) (net.Conn, HalfCloser, error) {
+	if !dialingFromServer {
+		return nil, nil, fmt.Errorf("must dial via mesh gateways from a server agent")
+	} else if gatewayResolver == nil {
+		return nil, nil, fmt.Errorf("gatewayResolver is nil")
+	} else if tlsConfigurator == nil {
+		return nil, nil, fmt.Errorf("tlsConfigurator is nil")
+	} else if dc == thisDatacenter {
+		return nil, nil, fmt.Errorf("cannot dial servers in the same datacenter via a mesh gateway")
+	} else if wrapper == nil {
+		return nil, nil, fmt.Errorf("cannot dial via a mesh gateway when outgoing TLS is disabled")
+	}
+
+	nextProto := actualRPCType.ALPNString()
+	if nextProto == "" {
+		return nil, nil, fmt.Errorf("rpc type %d cannot be routed through a mesh gateway", actualRPCType)
+	}
+
+	gwAddr := gatewayResolver(dc)
+	if gwAddr == "" {
+		return nil, nil, structs.ErrDCNotAvailable
+	}
+
+	dialer := &net.Dialer{LocalAddr: src, Timeout: defaultDialTimeout}
+
+	rawConn, err := dialer.Dial("tcp", gwAddr)
+	if err != nil {
+		return nil, nil, err
+	}
+
+	if tcp, ok := rawConn.(*net.TCPConn); ok {
+		_ = tcp.SetKeepAlive(true)
+		_ = tcp.SetNoDelay(true)
+	}
+
+	// NOTE: now we wrap the connection in a TLS client.
+	tlsConn, err := wrapper(dc, nodeName, nextProto, rawConn)
+	if err != nil {
+		return nil, nil, err
+	}
+
+	var conn net.Conn = tlsConn
+
+	var hc HalfCloser
+	if tlsConn, ok := conn.(*tls.Conn); ok {
+		// Expose *tls.Conn CloseWrite method on HalfCloser
+		hc = tlsConn
+	}
+
 	return conn, hc, nil
 }
 
 // getNewConn is used to return a new connection
-func (p *ConnPool) getNewConn(dc string, addr net.Addr, version int, useTLS bool) (*Conn, error) {
-	// Get a new, raw connection.
-	conn, _, err := p.DialTimeout(dc, addr, defaultDialTimeout, useTLS)
+func (p *ConnPool) getNewConn(dc string, nodeName string, addr net.Addr) (*Conn, error) {
+	if nodeName == "" {
+		return nil, fmt.Errorf("pool: ConnPool.getNewConn requires a node name")
+	}
+
+	// Get a new, raw connection and write the Consul multiplex byte to set the mode
+	conn, _, err := p.DialTimeout(dc, nodeName, addr, RPCMultiplexV2)
 	if err != nil {
 		return nil, err
 	}
 
-	// Switch the multiplexing based on version
-	var session muxSession
-	if version < 2 {
-		conn.Close()
-		return nil, fmt.Errorf("cannot make client connection, unsupported protocol version %d", version)
-	}
-
-	// Write the Consul multiplex byte to set the mode
-	if _, err := conn.Write([]byte{byte(RPCMultiplexV2)}); err != nil {
-		conn.Close()
-		return nil, err
-	}
-
-	// Setup the logger
 	conf := yamux.DefaultConfig()
-	conf.LogOutput = p.LogOutput
+	// override the default because LogOutput conflicts with Logger.
+	conf.LogOutput = nil
+	conf.Logger = p.Logger
 
 	// Create a multiplexed session
-	session, _ = yamux.Client(conn, conf)
+	session, err := yamux.Client(conn, conf)
+	if err != nil {
+		conn.Close()
+		return nil, fmt.Errorf("Failed to create yamux client: %w", err)
+	}
 
 	// Wrap the connection
 	c := &Conn{
 		refCount: 1,
+		nodeName: nodeName,
 		addr:     addr,
 		session:  session,
 		clients:  list.New(),
 		lastUsed: time.Now(),
-		version:  version,
 		pool:     p,
 	}
 	return c, nil
@@ -373,14 +489,19 @@ func (p *ConnPool) getNewConn(dc string, addr net.Addr, version int, useTLS bool
 
 // clearConn is used to clear any cached connection, potentially in response to an error
 func (p *ConnPool) clearConn(conn *Conn) {
+	if conn.nodeName == "" {
+		panic("pool: ConnPool.acquire requires a node name")
+	}
+
 	// Ensure returned streams are closed
 	atomic.StoreInt32(&conn.shouldClose, 1)
 
 	// Clear from the cache
 	addrStr := conn.addr.String()
+	poolKey := conn.nodeName + ":" + addrStr
 	p.Lock()
-	if c, ok := p.pool[addrStr]; ok && c == conn {
-		delete(p.pool, addrStr)
+	if c, ok := p.pool[poolKey]; ok && c == conn {
+		delete(p.pool, poolKey)
 	}
 	p.Unlock()
 
@@ -398,12 +519,12 @@ func (p *ConnPool) releaseConn(conn *Conn) {
 	}
 }
 
-// getClient is used to get a usable client for an address and protocol version
-func (p *ConnPool) getClient(dc string, addr net.Addr, version int, useTLS bool) (*Conn, *StreamClient, error) {
+// getClient is used to get a usable client for an address
+func (p *ConnPool) getClient(dc string, nodeName string, addr net.Addr) (*Conn, *StreamClient, error) {
 	retries := 0
 START:
 	// Try to get a conn first
-	conn, err := p.acquire(dc, addr, version, useTLS)
+	conn, err := p.acquire(dc, nodeName, addr)
 	if err != nil {
 		return nil, nil, fmt.Errorf("failed to get conn: %v", err)
 	}
@@ -425,11 +546,26 @@ START:
 }
 
 // RPC is used to make an RPC call to a remote host
-func (p *ConnPool) RPC(dc string, addr net.Addr, version int, method string, args interface{}, reply interface{}) error {
-	if method == "AutoEncrypt.Sign" {
-		return p.rpcInsecure(dc, addr, method, args, reply)
+func (p *ConnPool) RPC(
+	dc string,
+	nodeName string,
+	addr net.Addr,
+	method string,
+	args interface{},
+	reply interface{},
+) error {
+	if nodeName == "" {
+		return fmt.Errorf("pool: ConnPool.RPC requires a node name")
+	}
+
+	// TODO (autoconf) probably will want to have a way to invoke the
+	// secure or insecure variant depending on whether its an ongoing
+	// or first time config request. For now though this is fine until
+	// those ongoing requests are implemented.
+	if method == "AutoEncrypt.Sign" || method == "AutoConfig.InitialConfiguration" {
+		return p.rpcInsecure(dc, nodeName, addr, method, args, reply)
 	} else {
-		return p.rpc(dc, addr, version, method, args, reply)
+		return p.rpc(dc, nodeName, addr, method, args, reply)
 	}
 }
 
@@ -438,9 +574,13 @@ func (p *ConnPool) RPC(dc string, addr net.Addr, version int, method string, arg
 // transparent for the consumer. The pool cannot be used because
 // AutoEncrypt.Sign is a one-off call and it doesn't make sense to pool that
 // connection if it is not being reused.
-func (p *ConnPool) rpcInsecure(dc string, addr net.Addr, method string, args interface{}, reply interface{}) error {
+func (p *ConnPool) rpcInsecure(dc string, nodeName string, addr net.Addr, method string, args interface{}, reply interface{}) error {
+	if dc != p.Datacenter {
+		return fmt.Errorf("insecure dialing prohibited between datacenters")
+	}
+
 	var codec rpc.ClientCodec
-	conn, _, err := p.DialTimeoutInsecure(dc, addr, 1*time.Second, p.TLSConfigurator.OutgoingRPCWrapper())
+	conn, _, err := p.dial(dc, nodeName, addr, 0, RPCTLSInsecure)
 	if err != nil {
 		return fmt.Errorf("rpcinsecure error establishing connection: %v", err)
 	}
@@ -455,12 +595,11 @@ func (p *ConnPool) rpcInsecure(dc string, addr net.Addr, method string, args int
 	return nil
 }
 
-func (p *ConnPool) rpc(dc string, addr net.Addr, version int, method string, args interface{}, reply interface{}) error {
+func (p *ConnPool) rpc(dc string, nodeName string, addr net.Addr, method string, args interface{}, reply interface{}) error {
 	p.once.Do(p.init)
 
 	// Get a usable client
-	useTLS := p.TLSConfigurator.UseTLS(dc)
-	conn, sc, err := p.getClient(dc, addr, version, useTLS)
+	conn, sc, err := p.getClient(dc, nodeName, addr)
 	if err != nil {
 		return fmt.Errorf("rpc error getting client: %v", err)
 	}
@@ -490,9 +629,9 @@ func (p *ConnPool) rpc(dc string, addr net.Addr, version int, method string, arg
 
 // Ping sends a Status.Ping message to the specified server and
 // returns true if healthy, false if an error occurred
-func (p *ConnPool) Ping(dc string, addr net.Addr, version int) (bool, error) {
+func (p *ConnPool) Ping(dc string, nodeName string, addr net.Addr) (bool, error) {
 	var out struct{}
-	err := p.RPC(dc, addr, version, "Status.Ping", struct{}{}, &out)
+	err := p.RPC(dc, nodeName, addr, "Status.Ping", struct{}{}, &out)
 	return err == nil, err
 }
 
