@@ -8,14 +8,13 @@ import (
 var (
 	// The git commit that was compiled. These will be filled in by the
 	// compiler.
-	GitCommit   string
-	GitDescribe string
+	GitCommit string
 
 	// The main version number that is being run at the moment.
 	//
 	// Version must conform to the format expected by github.com/hashicorp/go-version
 	// for tests to work.
-	Version = "1.7.4"
+	Version = "1.8.6"
 
 	// A pre-release marker for the version. If this is "" (empty string)
 	// then it means that it is a final release. Otherwise, this is a pre-release
@@ -27,23 +26,13 @@ var (
 // for displaying to humans.
 func GetHumanVersion() string {
 	version := Version
-	if GitDescribe != "" {
-		version = GitDescribe
-	}
-
 	release := VersionPrerelease
-	if GitDescribe == "" && release == "" {
-		release = "dev"
-	}
 
 	if release != "" {
 		if !strings.HasSuffix(version, "-"+release) {
 			// if we tagged a prerelease version then the release is in the version already
 			version += fmt.Sprintf("-%s", release)
 		}
-		if GitCommit != "" {
-			version += fmt.Sprintf(" (%s)", GitCommit)
-		}
 	}
 
 	// Strip off any single quotes added by the git information.
